#include <Inventor/nodes/SoSeparator.h>
#include <Inventor/nodes/SoSelection.h>

/*
   Should not be here...
*/
#include <Inventor/nodes/SoMaterial.h>
#include <Inventor/nodes/SoLineSet.h>
#include <Inventor/nodes/SoCoordinate3.h>

/*
 Qt Components
*/
#include <Inventor/Qt/SoQt.h>
#include <Inventor/Qt/SoQtRenderArea.h>
#include <Inventor/Qt/viewers/SoQtExaminerViewer.h>
#include <Inventor/SbViewportRegion.h>
#include <Inventor/nodes/SoDirectionalLight.h>
#include <Inventor/nodes/SoLightModel.h>

#include <Inventor/manips/SoHandleBoxManip.h>
#include <Inventor/manips/SoTrackballManip.h>

#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/actions/SoGetBoundingBoxAction.h>

#include <Inventor/actions/SoWriteAction.h>
#include <Inventor/SoOutput.h>

#include <stdlib.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <fcntl.h>

#include <assert.h>
#include "ModelInvPts.h"
#include "CoordSystem.h"
#include "Axis.h"

SoQtExaminerViewer *MainNpExaminerViewer;  
SoSeparator *SceneRoot;  
Model *_theModel;
M_CoordSystem *CoordS=NULL;

static int _save_iv_file=0;

int  read_seq_of_pointsCYRA(Model *&theModel, FILE *fpts, int MaxNum);
void put_lights            (SoSeparator *sceneRoot);
void createCoordSystem     (SoSeparator *root);


FILE *_f_pts;

//Set this flag to 1 if you want to see all points in one color
//Set this flag to 0 if you want to see points in scan line order
//const int _FLAG_SHOW_ALL=0;

main(int argc, char **argv){


  int from_scan_line;
  int to_scan_line;
    if (argc!=4) {
	printf("Usage: %s  input_ptx_file_name from_scan_line to_scan_line\n",argv[0]);
	printf("      if from and to are both 0 everything is being displayed\n");
	return(0);
    }

    from_scan_line=atoi(argv[2]);
    to_scan_line  =atoi(argv[3]);

   if (from_scan_line==0 && to_scan_line==0){
     printf("Will show all scan lines\n");
    } 
    
    struct timeval tp;
    gettimeofday(&tp,NULL);
    srand48( (long) tp.tv_sec );

    int nrows, ncols;


    QWidget *myWindow = SoQt::init(argv[0]);
    if (myWindow == NULL) exit(1);



    SoQtExaminerViewer *MainExaminerViewer=new SoQtExaminerViewer( myWindow );
    SceneRoot = new SoSeparator;
    SceneRoot->ref();
    put_lights(SceneRoot);



    if ( (_f_pts=fopen(argv[1], "r"))==NULL ){
      printf("Can't open file %s for reading\n", argv[1]);
      return 1;
    }
    
    //Read Header
    fscanf(_f_pts, "%d\n", &nrows);
    //File can not have negative or zero number of rows
    if (nrows<=0) {
	fprintf(stderr, "range_data::read_data_ascii-> invalid num of rows\n");
	return 1;
    }
    fscanf(_f_pts, "%d\n", &ncols);
    //File can not have negative or zero number of columns
    if (ncols<=0) {
	fprintf(stderr, "range_data::read_data_ascii-> invalid num of cols\n");
	return 1;
    }
    //Print to  stdout for information purposes
    printf("read_data-> Reading %d rows and %d cols\n", nrows, ncols);


    double dummy[4];
    //Just skip the first four lines (no useful information)
    for (int i=0; i<4; ++i){
      fscanf(_f_pts, "%lf %lf %lf\n", &dummy[0], &dummy[1], &dummy[2]);
    }
    //The remaining four lines contain the rotation matrix / translation vector
    //Skip
    for (int i=0; i<4; ++i){
      fscanf(_f_pts, "%lf %lf %lf %lf\n", &dummy[0], &dummy[1], &dummy[2], &dummy[3]);
    }
    
    int more_to_read=0;

    int num_of_pts_per_iteration;
//     if (_FLAG_SHOW_ALL) num_of_pts_per_iteration=nrows*ncols;
//     else num_of_pts_per_iteration=ncols;

    //if (scan_line_to_show == -1) num_of_pts_per_iteration=nrows*ncols;
    //else num_of_pts_per_iteration=ncols;

    num_of_pts_per_iteration=ncols;

//     if (scan_line_to_show!=-1) {
//     if (scan_line_to_show <=0 || scan_line_to_show > nrows) {
//       printf("Scan line %d is out of bounds.. Range should be from 1 to %d\n",
// 	     scan_line_to_show, nrows);
//       return 0;
//     }
//     }

    if (from_scan_line!=0 || to_scan_line!=0){
      if (from_scan_line <=0 || from_scan_line > nrows || to_scan_line <=0 || to_scan_line > nrows){
	printf("Scan lines out of range\n");
	return 1;
      }
      if (from_scan_line > to_scan_line){
	printf("From scan line can't be after to_scan_line\n");
	return 1;
      }
    }

    int current_row=0;
    do {
      Model *theModel=new Model;
      assert( theModel != NULL );
      theModel->RandomColor();
      more_to_read=read_seq_of_pointsCYRA(theModel, _f_pts, num_of_pts_per_iteration);
      //      if (hack==nrows-1){

      //      if (current_row==scan_line_to_show-1 || scan_line_to_show==-1){
      if ( (from_scan_line ==0 && to_scan_line==0) ||
	   current_row >= from_scan_line-1 && current_row <= to_scan_line-1 ){
	SceneRoot->ref();
	SceneRoot->addChild( theModel->getRoot() );
      }
      //  }
      //      hack++;
      //    }while (hack < nrows);
      current_row ++;
    } while (more_to_read);

    createCoordSystem(SceneRoot);
    fclose(_f_pts);




    MainExaminerViewer->setSceneGraph( SceneRoot );
    MainExaminerViewer->show();
    SoQt::show(myWindow);

    if (!_save_iv_file){
      SoWriteAction *myAction=new SoWriteAction;
      myAction->getOutput()->openFile("TheScene.iv");
      myAction->getOutput()->setBinary(FALSE);
      myAction->apply( SceneRoot );
      myAction->getOutput()->closeFile();
      _save_iv_file=1;
    }   	    


    SoQt::mainLoop();

}


//Small module
const int NUM_PTS=1000000;
static float _PtsArray[NUM_PTS][3];
int
read_seq_of_pointsCYRA(Model *&theModel, FILE *fpts, int maxNumOfPoints )
{
  float x, y, z, dump;
  int flagEOF=0;
  int _num_pts=0;

  //int pNum=-1;

  while ( _num_pts<maxNumOfPoints && !flagEOF ) {
    if (fscanf(fpts, "%f %f %f %f", &x, &y, &z, &dump)!=EOF){
      //pNum++;
      if (_num_pts >= NUM_PTS-1 ) {
	printf("Too many points...\n");
	theModel->addPoints( _PtsArray, _num_pts);
	return 1;
      }
      //if (pNum>=243 && pNum <251){
      _PtsArray[_num_pts][0] = x;
      _PtsArray[_num_pts][1] = y;
      _PtsArray[_num_pts][2] = z;
      ++_num_pts; 
      //}
    } else flagEOF=1;
  }
  
  theModel->addPoints(_PtsArray, _num_pts);
  if (flagEOF) return 0; //no more points
  else return 1;
}
//End small module



void
put_lights(SoSeparator *sceneRoot)
{
//    SoSeparator *LightRoot = new SoSeparator;

    SoDirectionalLight *mydLight=new SoDirectionalLight;
    mydLight->direction.setValue(1, 1.5, 0.5);

    SoDirectionalLight *mypLight=new SoDirectionalLight;
    mypLight->direction.setValue( -0.7, -0.4 , 0.1 );

    SoDirectionalLight *myxLight=new SoDirectionalLight;
    mypLight->direction.setValue( 0.3, -0.5, 2.0 );

    sceneRoot->addChild(mydLight);
    sceneRoot->addChild(mypLight);
    sceneRoot->addChild(myxLight);

    SoLightModel *myModel=new SoLightModel;
    myModel->model = SoLightModel::PHONG;
    sceneRoot->addChild(myModel);

 //   return LightRoot;
}



//Copied from NewModel.C
void
createCoordSystem (SoSeparator *root) {

  printf("create coord system...\n");
 float smax;
  /*  
SbViewportRegion VpRegion = MainNpExaminerViewer->getViewportRegion();
  SoGetBoundingBoxAction *BBaction =
      new  SoGetBoundingBoxAction( VpRegion );
  assert(BBaction!=0);



  BBaction->apply( root );


  float sx, sy, sz;
 
  SbBox3f bbox = BBaction->getBoundingBox();

  bbox.getSize(sx, sy, sz);
  printf("size= %f %f %f\n", sx, sy, sz);

  if (sx > sy) smax=sx;
     else smax=sy;
  if (sz > smax) smax=sz;

  */
  smax=50;
  printf("max: %f\n", smax);
  if (CoordS==NULL){
      CoordS= new M_CoordSystem( smax, 10);
      assert(CoordS!=0);
    }
  printf("Now I will create the axis...\n");
  M_Axis *m_Axis= new M_Axis;
  assert(m_Axis!=0);
  SoSeparator *axisSep=m_Axis->get_Root();
  printf("I got the coord. sys. root...\n");
  root->addChild( axisSep );

  // delete BBaction;
}


